/*
 ***************************************************************************
 *
 * Copyright 2004 Michael Mueller, michael.mueller@kay-mueller.de
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA. 
 *
 ***************************************************************************
 *
 * Log network connections using /proc/net/ip_conntrack or nf_conntrack.
 * This only works under Linux 2.4 and up.
 *
 * usage: ipclog [options]
 *     options:
 *        -d               : run in foreground
 *        -c config-file   : config file
 *        -h               : help
 *        -o output-file   : send log output to output-file
 *        -t time-interval : for conntrack lookups (secs)
 *        -I               : show icmp entries
 *        -i               : don't show if name instead of addr
 *        -N               : show NAT info
 *        -1               : show open connections and exit
 *
 * Linking ipclog to 'ipc' and calling it is equivalent to 'ipclog -1'
 * which shows currently open connections and exits.
 *
 * Log file format:
 *
 *  The character in the first column indicates the type of entry. Each 
 *  type has it's own format. The second column always shows date and time.
 *
 *  Type 'P':
 *    programm message, like "P 2004-07-24-14:10:52 ipclog started"
 *
 *  Type 'H':
 *    column header for 'O', 'C', 'c', 'S' types: 
 *     "H 2004-07-24-14:10:52 prot src dst sport dport fl time"
 *
 *  Types 'O', 'C', 'c', 'S':
 *    'O' connection opende
 *    'C' connection closed
 *    'c' connection open at program exit
 *    'S' tcp state change (for example into TIME_WAIT state)
 *
 *        column 1: 'O', 'C', 'c', 'S'
 *	column 2: date and time
 *	column 3: protocol
 *	column 4: source address, my be name of local interface
 *	column 5: destination address, my be name of local interface
 *	column 6: source port
 *	column 7: destination port
 *	column 8: flags
 *	  Flag values:
 *	    'U' - UNREPLIED
 *	    'a' - tcp, LAST_ACK
 *	    'e' - tcp, ESTABLISHED
 *	    'f' - tcp, FIN_WAIT
 *	    'r' - tcp, SYN_RECV
 *	    's' - tcp, SYN_SENT
 *	    't' - tcp, TIME_WAIT
 *	    'w' - tcp, CLOSE_WAIT
 *	column 9: connection time in seconds
 *
 *	If the -N option (show NAT info) is used, additional columns are:
 *
 *	column 10: snat, source nat address
 *	column 11: dnat, destination nat address
 *	column 12: sn_po, source nat port
 *	column 13: dn_po, destination nat port
 *
 *  Type 'I':
 *    network interface up, like "I 2004-07-24-14:10:52 UP eth0 192.168.0.1"
 *
 *  Type 'i':
 *    network interface down, like "i 2004-07-24-14:10:52 DOWN eth0 192.168.0.1"
 *
 *  Type 'E':
 *    error message
 *
 * Configuration file:
 *
 *   Default path is /etc/ipclogd.conf. See also -c option. Useful for
 *   specifying alias names for multiple addresses of the same network 
 *   interface. (For example ppp0 with additional address for vpn network.)
 *
 *   Example:
 *
 *	# this is a comment
 *	ifname ppp0 192.132.3.0/24 vpn1               # vpn 1
 *	ifname ppp0 192.132.4.0/24 vpn2               # vpn 2
 *	# debug                                       # equivalent to -d option
 *
 *
 *
 ***************************************************************************
 */

#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <net/if.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <time.h>
#include <unistd.h>
#include <syslog.h>
#include <errno.h>
#include <signal.h>
#include <sys/ioctl.h>

#define IP_CONNTRACK_FILE "/proc/net/nf_conntrack"
#define IP_CONNTRACK_FILE_OLD "/proc/net/ip_conntrack"
int ctFileVersion = 2; /* 1 = nf_conntrack, 2 = nf_conntrack */

struct ipConntrackEntry {
	u_short protocol;
	unsigned char flags;
	unsigned char intFlags;
	time_t timeout;
	time_t openTime;
	struct in_addr src[2];
	int srcIfIdx[2]; /* index into ifList */
	struct in_addr dst[2];
	int dstIfIdx[2]; /* index into ifList */
	union {
		struct {
			u_short state;	/* tcp only */
			u_short sport[2];
			u_short dport[2];
		} tcp_udp;
		struct {
			u_int8_t type[2];
			u_int8_t code[2];
			u_int16_t id[2];
		} icmp;
	} u;
	time_t lastActive;
	struct ipConntrackEntry *next;
};

/* ipConntrackEntry.flags */
#define F_ASSURED	(1<<0)
#define F_UNREPLIED	(1<<1)

/* ipConntrackEntry.intFlags */
#define FI_SEEN		(1<<0)
#define FI_IGNORE	(1<<1)

struct {
	char *lName;
	char *sName;
} tcp_states[] = {
	{ "NONE",	"n" },	/* 0 */
	{ "ESTABLISHED","e" },	/* 1 */
	{ "SYN_SENT",	"s" },	/* 2 */
	{ "SYN_RECV",	"r" },	/* 3 */
	{ "FIN_WAIT",	"f" },	/* 4 */
	{ "TIME_WAIT",	"t" },	/* 5 */
	{ "CLOSE",	"c" },	/* 6 */
	{ "CLOSE_WAIT",	"w" },	/* 7 */
	{ "LAST_ACK",	"a" },	/* 8 */
	{ "LISTEN",	"l" },	/* 9 */
	{ NULL,		NULL}
};

#define TCP_FIN_WAIT  	4
#define TCP_TIME_WAIT	5
#define TCP_CLOSE 	6
#define TCP_CLOSE_WAIT 	7
#define TCP_LAST_ACK 	8

#define TCP_ENDSTATE(s) ((s) >= TCP_FIN_WAIT && (s) <= TCP_LAST_ACK)

/* List of network interfacees we have seen so far. The name is the
 * interface's real name. The nextSeqNum is increased every time
 * we find the same interface with a new address. (Intefaces can
 * have more than one address assigned.)
 */
struct ifNames {
	char name[IF_NAMESIZE+1];
	unsigned nextSeqNum;
} *ifNames;

int numIfNames = 0;

/* List of network interface / address pairs.
 * The interface's real name is ifNames[nameIdx].name.
 * Entries are referenced by index. The name is
 * generated from ifNames[nameIdx].name - ifNames[nameIdx].nextSeqNum,
 * for example "ppp0-3". But instead of "ppp0-0" simply "ppp0" is used.
 */
struct ifList {
	int nameIdx;			/* index into ifNames[] */
	char name[IF_NAMESIZE+11+1];	/* generated from ifNames[].name */
	struct in_addr addr;		/* inet address			 */
	unsigned seqNum;		/* sequence number from ifNames	 */
	int isUp;			/* interface is up		 */
	int oldIsUp;			/* old value of isUp up		 */
} *ifList;

int numIf = 0;

#define IF_IDX_NONE	(-1)

/* 
 * Interface name map read from config file
 * Maps name with address addr/mask to newName.
 */
#define IF_MAP_MAX 40
struct ifMap {
	char name[IF_NAMESIZE+1];
	char newName[IF_NAMESIZE+1];
	struct in_addr addr;
	unsigned mask;
} ifMap[IF_MAP_MAX];

int ifMapLen = 0;

int termFlag = 0; /* terminate prog */

int syslog_open = 0;
int dFlag = 0;
int nFlag = 0;
int icmpFlag = 0;
int tinterval = 5;
int onceFlag = 0;
int noTcpTermConFlag = 0;	/* don't log terminating TCP connections */
int noIfFlag = 0;
FILE *outFp;

int errCnt = 0;

struct ipConntrackEntry * entryCache = NULL;

#define MAX_CONN_HASH 1024 
struct ipConntrackEntry *connHash[MAX_CONN_HASH];

/* log types */
#define LOG_ERROR	"E"
#define LOG_HEADER	"H"
#define LOG_OPEN_CONN	"O"
#define LOG_CLOSE_CONN	"C"
#define LOG_EXIT_CONN	"c"	/* connection open at prog exit */
#define LOG_TCP_STATE	"S"
#define LOG_IF_UP	"I"
#define LOG_IF_DOWN	"i"
#define LOG_PROGRAM	"P"

#define SKIP_BLANK(p) {				\
	while (*(p) == ' ')			\
		++(p);				\
}

#define SKIP_NON_BLANK(p) {			\
	while (*(p) != '\0' && *(p) != ' ')	\
		++(p);				\
}



void Error(char *str, ...);
void Usage(char *prog);
void ParseCfgFile(char *cfgfile);
void Log(char *type, time_t t, char *msg, ...);
void PrintEntry(struct ipConntrackEntry *entry);
void LogEntry(char *type, time_t t, struct ipConntrackEntry *entry);
void SetTermFlag(int sig);
void ProcessIpConntrack(void);
char *ReadIpConntrack(int *len);
void ReadIfList(time_t currTime);
int GetIfIndex(struct in_addr addr);
int GetIf(char *name, struct in_addr addr);
void RemapIf(char *name, struct in_addr addr);
int InsertIf(char *name, struct in_addr addr);
void AddrToAscii(char *s, int len, struct in_addr addr, int ifIdx);
char *ScanIpConntrack(char *p, struct ipConntrackEntry *entry, int *err);
char *ScanTcp(char *p, struct ipConntrackEntry *entry);
char *ScanTuple(char *p, int n, struct ipConntrackEntry *entry, char *ps);
char *ScanTupleTcpUdp(char *p, int n, struct ipConntrackEntry *entry, char *ps);
char *ScanTupleIcmp(char *p, int n, struct ipConntrackEntry *entry, char *ps);
struct ipConntrackEntry *AllocEntry();
void FreeEntry(struct ipConntrackEntry *e);
unsigned Hash(struct ipConntrackEntry *e);
struct ipConntrackEntry * FindInsert(struct ipConntrackEntry *e);
int Equal(struct ipConntrackEntry *e1, struct ipConntrackEntry *e2);


int
main(int argc, char **argv)
{
	int n, c, e, fd;
	char *outfile = "-", *prog;
	char def_cfgfile[] = { "/etc/ipclogd.conf" };
	char *cfgfile = def_cfgfile;
	time_t currTime;
	struct ipConntrackEntry *entry;

	prog = rindex(argv[0], '/');
	if (!prog)
		prog = argv[0];
	else
		++prog;

	if (!strcmp(prog, "ipc")) {
		onceFlag = 1;
		noTcpTermConFlag = 1;
		dFlag = 1;
	}


	while ((c = getopt(argc, argv, ":1INdhic:o:t:")) != -1)
		switch (c) {
			case '1':
				onceFlag = 1;
				noTcpTermConFlag = 1;
				dFlag = 1;
				break;
			case 'I':
				icmpFlag = 1;
				break;
			case 'N':
				nFlag = 1;
				break;
			case 'c':
				cfgfile = optarg;
				break;
			case 'd':
				dFlag = 1;
				break;
			case 'h':
				Usage(prog);
			case 'i':
				noIfFlag = 1;
				break;
			case 'o':
				outfile = optarg;
				break;
			case 't':
				tinterval = atoi(optarg);
				break;
			case ':':
				Error("option -%c requires an argument", 
					optopt);
			case '?':
				Error("unrecognized option: -%c", optopt );
			default:
				Error("bad option");
		}

	if (optind != argc)
		Usage(prog);

	ParseCfgFile(cfgfile);

	if (!dFlag) {
		if ((e = fork()) < 0)
			Error("fork");
		if (e > 0)
			exit(0);

		fd = open("/dev/null", O_RDWR);
		dup2(fd, 0);
		if (strcmp(outfile, "-"))
			dup2(fd, 1);
		dup2(fd, 2);
		close(fd);

		openlog("ipclog", LOG_PID, LOG_DAEMON);
		if (*cfgfile)
			syslog(LOG_NOTICE, "ipclog started, cfgfile %s\n",
				cfgfile);
		else
			syslog(LOG_NOTICE, "ipclog started, no cfgfile\n");
		syslog_open = 1;

		if (setsid() < 0)
			Error("setsid");
	}

	signal(SIGTERM, SetTermFlag);
	signal(SIGINT, SetTermFlag);
	signal(SIGHUP, SIG_IGN);

	if (!strcmp(outfile, "-"))
		outFp = stdout;
	else {
		outFp = fopen(outfile, "a");
		if (!outFp)
			Error("cannot open file %s", outfile);
	}

	if (dFlag && !onceFlag)
		fprintf(outFp, "ipclog started\n");

	currTime = time(NULL);
	if (!onceFlag)
		Log(LOG_PROGRAM, currTime, "ipclog started");

	if (nFlag) {
	    Log(LOG_HEADER, currTime, "%-4s %-15s %-15s %-5s %-5s %-2s %-5s "
	    	"%-15s %-15s %5s %5s",
		"prot", "src", "dst", "sport", "dport", "fl", " time",
		"snat", "dnat", "sn_po", "dn_po");
	    if (icmpFlag)
	      Log(LOG_HEADER, currTime, "%-4s %-15s %-15s %-5s %-5s %-2s %-5s "
	    	    "%-15s %-15s %5s %5s",
		    "icmp", "src", "dst", "type", "code", "fl", " time",
		    "snat", "dnat", "-", "-");
	}
	else {
	    Log(LOG_HEADER, currTime, "%-4s %-15s %-15s %-5s %-5s %-2s %-5s",
		"prot", "src", "dst", "sport", "dport", "fl", " time");
	    if (icmpFlag)
	      Log(LOG_HEADER, currTime, "%-4s %-15s %-15s %-5s %-5s %-2s %-5s",
		    "icmp", "src", "dst", "type", "code", "fl", " time");
	}

	setvbuf(outFp, NULL, _IOLBF, 240);

	while (!termFlag) {
		ProcessIpConntrack();
		if (onceFlag)
			exit(0);

		if (termFlag)
			break;

		sleep(tinterval);
	}

	currTime = time(NULL);
	for (n = 0; n < MAX_CONN_HASH; ++n) {
		for (entry = connHash[n]; entry; entry = entry->next)
			LogEntry(LOG_EXIT_CONN, currTime, entry);
	}

	Log(LOG_PROGRAM, currTime, "ipclog stopped, %d errors", errCnt);
	if (!dFlag) {
		syslog(LOG_NOTICE, "ipclog stopped, %d errors\n", errCnt);
		closelog();
	}
	else
		fprintf(outFp, "ipclog stopped, %d errors\n", errCnt);

	exit(0);

	return 0;
}

void
ProcessIpConntrack()
{
	char *p, *currLine, *pEnd;
	int n, l, err;
	struct ipConntrackEntry *entry, *e, *prev, *next;
	time_t currTime, t;

	currTime = time(NULL);

	if (!(entry = AllocEntry())) {
		Log(LOG_ERROR, currTime, "out of memory");
		return;
	}


	p = ReadIpConntrack(&l);
	pEnd = p + l;

	if (!noIfFlag)
		ReadIfList(currTime);

	/* scan conntrack pseudo file and put connections into hash table */
	while (p && *p) {
		currLine = p;
		p = ScanIpConntrack(p, entry, &err);
		if (err)
			continue;

		/* PrintEntry(entry); */

		if (entry->protocol == IPPROTO_ICMP && !icmpFlag)
			continue;

		e = FindInsert(entry);

		if (e == entry) { /* new entry */
			e->timeout += currTime;
			e->openTime = currTime;
			e->lastActive = currTime;

			e->srcIfIdx[0] = GetIfIndex(e->src[0]);
			e->srcIfIdx[1] = GetIfIndex(e->src[1]);
			e->dstIfIdx[0] = GetIfIndex(e->dst[0]);
			e->dstIfIdx[1] = GetIfIndex(e->dst[1]);

			/* if noTcpTermConFlag don't log tcp time wait etc. 
			 * first time through */
			if (noTcpTermConFlag && e->protocol == IPPROTO_TCP && 
			    TCP_ENDSTATE(e->u.tcp_udp.state))
				e->intFlags |= FI_IGNORE;
			else
				LogEntry(LOG_OPEN_CONN, currTime, entry);

			if (!(entry = AllocEntry())) {
				Log(LOG_ERROR, 0, "out of memory");
				return;
			}
		}
		else { /* old entry, update it */
			e->flags = entry->flags;
			t = currTime + entry->timeout;

			/* If timeout value changed significantly (+- 1 sec)
			 * it must be due to activity.
			 */
			if (t < e->timeout - 1 || t > e->timeout + 1) {
				e->timeout = t;
			    	e->lastActive = currTime;
			}
			if (entry->protocol == IPPROTO_TCP) {
				u_short state = e->u.tcp_udp.state;
				e->u.tcp_udp.state = entry->u.tcp_udp.state;
				if (e->u.tcp_udp.state != state)
					LogEntry(LOG_TCP_STATE, currTime, e);
			}
		}

		e->intFlags |= FI_SEEN;
	}

	if (p != pEnd)
		Error("internal error in ProcessIpConntrack: p != pEnd");

	FreeEntry(entry);

	/* log and delete all entries that have disappeared from 
	 * conntrack file 
	 */
	for (n = 0; n < MAX_CONN_HASH; ++n) {
		for (e = connHash[n], prev = NULL; e; e = next) {
			next = e->next;

			if (e->intFlags & FI_SEEN) {
				e->intFlags &= ~FI_SEEN;
				prev = e;
				continue;
			}

			LogEntry(LOG_CLOSE_CONN, e->lastActive, e);

			if (!prev)
				connHash[n] = e->next;
			else
				prev->next = e->next;
			FreeEntry(e);
		}
	}
}

/*
 * The read call for pseudo file /proc/net/xx_conntrack behaves a bit
 * unusual: It always returns n whole lines. If the first line doesn't fit
 * into the buffer, it returns 0, which is NOT EOF, but "buffer to small"!
 *
 * We make sure our buffer space is at least MAX_LINE_SIZE for each read call.
 * If the whole file doesn't fit, we increase the buffer size and retry.
 */
char *
ReadIpConntrack(int *len)
{
#	define MIN_BUF_SIZE (8*1024)
#	define MAX_LINE_SIZE 1024	/* should be enough ?! */
#	define BUF_INC(sz) ((sz) * 12 / 10)

	static char *p, *buf = NULL;
	static int bufsize = MIN_BUF_SIZE;

	int fd, n = 0, l, firstTime = 1;

	if (!buf) {
		buf = malloc(bufsize);
		if (!buf)
			Error("cannot malloc first buf");
	}

	do {
		if (!firstTime) {
			bufsize = BUF_INC(bufsize);
			buf = realloc(buf, bufsize);
			if (!buf)
				Error("cannot realloc buf");
			}
		firstTime = 0;

		fd = open(IP_CONNTRACK_FILE, O_RDONLY);
		if (fd < 0) {
			fd = open(IP_CONNTRACK_FILE_OLD, O_RDONLY);
			if (fd < 0)
				Error("cannot open %s or %s", 
				    IP_CONNTRACK_FILE, IP_CONNTRACK_FILE_OLD);
			ctFileVersion = 1;

		}
		p = buf;
		l = 0; 
		do {
			if (l >= bufsize - MAX_LINE_SIZE)
				break;

			n = read(fd, p, bufsize - 1 - l);
			if (n < 0)
				Error("cannot read conntrack file");
			l += n;
			p += n;
		} while (n > 0);

		close(fd);
	} while (n > 0);

	buf[l] = 0;
	*len = l;
	return buf;
}

/* 
 * Get the up interfaces from the system and update ifList
 */

void 
ReadIfList(time_t currTime)
{
	static char *buf = NULL;
	static int bufLen = 0;

	int sd;
	struct ifconf ifc;
	struct ifreq *ifr;
	int num, n, ifIdx;
	struct sockaddr_in *saddr;
	char if_name[IF_NAMESIZE+1];

	if (noIfFlag)
		return;

	sd = socket(AF_INET, SOCK_DGRAM, 0);
	if (sd < 0)
		Error("ReadIfList:socket");

	ifc.ifc_len = 0;
	ifc.ifc_buf = NULL;
	if (ioctl(sd, SIOCGIFCONF, (char *) &ifc) < 0)
		Error("ReadIfList:ioctl1 SIOCGIFCONF");

	if (!bufLen && !ifc.ifc_len) {
		close(sd);
		return;
	}

	if (bufLen < ifc.ifc_len) {
		bufLen = ifc.ifc_len;
		buf = realloc(buf, bufLen);
		if (!buf)
			Error("ReadIfList:realloc");
	}
	ifc.ifc_buf = (caddr_t) buf;
	if (ioctl(sd, SIOCGIFCONF, (char *) &ifc) < 0)
		Error("ReadIfList:ioctl2 SIOCGIFCONF");

	close(sd);

	num = ifc.ifc_len / sizeof(struct ifreq);

	/* move ifList[].isUp into oldIsUp */
	for (n = 0; n < numIf; ++n) {
		ifList[n].oldIsUp = ifList[n].isUp;
		ifList[n].isUp = 0;
	}

	/* update / insert ifList */
	for (n = 0, ifr = ifc.ifc_req; n < num; ++n, ++ifr) {
		strncpy(if_name, ifr->ifr_name, IF_NAMESIZE);
		if_name[IF_NAMESIZE] = 0;
		saddr = (struct sockaddr_in *)&ifr->ifr_addr;
		RemapIf(if_name, saddr->sin_addr);

		ifIdx = GetIf(if_name, saddr->sin_addr);
		if (ifIdx < 0)
			ifIdx = InsertIf(if_name, saddr->sin_addr);
		ifList[ifIdx].isUp = 1;
	}

	if (onceFlag)
		return;

	/* print out changes */
	for (n = 0; n < numIf; ++n) {
		if (ifList[n].isUp == ifList[n].oldIsUp)
			continue;

		if (ifList[n].oldIsUp)
			Log(LOG_IF_DOWN, currTime, "DOWN %-13s   %-15s",
				ifList[n].name, inet_ntoa(ifList[n].addr));

		if (ifList[n].isUp)
			Log(LOG_IF_UP,   currTime, "UP   %-13s   %-15s",
				ifList[n].name, inet_ntoa(ifList[n].addr));
	}
}

/*
 * Get index in ifList for interface with real name name and address addr.
 *
 */

int 
GetIf(char *name, struct in_addr addr)
{
	int n, i;

	for (n = 0; n < numIf; ++n) {
		i = ifList[n].nameIdx;
		if (!strcmp(ifNames[i].name, name) &&
		    ifList[n].addr.s_addr == addr.s_addr)
			return n;
	}

	return IF_IDX_NONE;
}

/*
 * Get index in ifList for interface with address addr.
 *
 */

int 
GetIfIndex(struct in_addr addr)
{
	int n;

	for (n = 0; n < numIf; ++n)
		if (ifList[n].addr.s_addr == addr.s_addr)
			return n;

	return IF_IDX_NONE;
}

/*
 * Insert new up interface into ifList. Also insert / update ifNames.
 */

int 
InsertIf(char *name, struct in_addr addr)
{
	int ifNameIdx;

	/* insert / update ifNames */
	for (ifNameIdx = 0; ifNameIdx < numIfNames; ++ifNameIdx)
		if (!strcmp(ifNames[ifNameIdx].name, name))
			break;
	if (ifNameIdx == numIfNames) { /* insert into ifNames */
		ifNames = realloc(ifNames, 
			(numIfNames + 1) * sizeof(struct ifNames));
		if (!ifNames)
			Error("InsertIf: realloc %d", numIfNames + 1);
		strncpy(ifNames[numIfNames].name, name, IF_NAMESIZE+1);
		ifNames[numIfNames].name[IF_NAMESIZE] = 0;
		ifNames[numIfNames].nextSeqNum = 0;
		ifNameIdx = numIfNames++;
	}
	/* interface is in ifNames[ifNameIdx] */

	/* realloc ifList */
	ifList = realloc(ifList, (numIf + 1) * sizeof(struct ifList));
	if (!ifList)
		Error("InsertIf: realloc %d", numIf + 1);

	ifList[numIf].nameIdx = ifNameIdx;
	ifList[numIf].addr = addr;
	ifList[numIf].isUp = 1;
	ifList[numIf].oldIsUp = 0;

	if (ifNames[ifNameIdx].nextSeqNum > 0)
		sprintf(ifList[numIf].name, "%s-%u", ifNames[ifNameIdx].name, 
			ifNames[ifNameIdx].nextSeqNum);
	else {
		strncpy(ifList[numIf].name, ifNames[ifNameIdx].name,
			IF_NAMESIZE+1);
		ifList[numIf].name[IF_NAMESIZE] = 0;
	}
	ifList[numIf].seqNum = ifNames[ifNameIdx].nextSeqNum++;

	return numIf++;
}

void
AddrToAscii(char *s, int len, struct in_addr addr, int ifIdx)
{
	if (ifIdx == IF_IDX_NONE) {
		strncpy(s, inet_ntoa(addr), len);
		s[len-1] = 0;
	}
	else {
		if (ifIdx < 0 || ifIdx >= numIf)
			Error("AddrToAscii: internal error");
		strncpy(s, ifList[ifIdx].name, len);
		s[len-1] = 0;
	}
}

char *
ScanIpConntrack(char *p, struct ipConntrackEntry *entry, int *err)
{
	char *ps, *pe;
	int n;

	*err = 1;
	memset(entry, 0, sizeof(struct ipConntrackEntry));
	entry->srcIfIdx[0] = IF_IDX_NONE;
	entry->srcIfIdx[1] = IF_IDX_NONE;
	entry->dstIfIdx[0] = IF_IDX_NONE;
	entry->dstIfIdx[1] = IF_IDX_NONE;

	ps = p;
	pe = index(p, '\n');
	if (!pe) {
		Log(LOG_ERROR, 0, "scan error 1: %s", ps);
		return NULL;
	}
	*pe++ = 0;

	if (ctFileVersion == 1) {
		if (sscanf(p, "%*s%hu%u%n", &entry->protocol, 
						&entry->timeout, &n) != 2) {
			Log(LOG_ERROR, 0, "scan error 2: %s", ps);
			return pe;
		}
	}
	else {
		unsigned proto;
		if (sscanf(p, "%*s%u%*s%hu%u%n", &proto, &entry->protocol, 
				&entry->timeout, &n) != 3 || proto != 2) {
			Log(LOG_ERROR, 0, "scan error 2: %s", ps);
			return pe;
		}
	}
	p += n;

	if (entry->protocol == IPPROTO_TCP) {
		p = ScanTcp(p, entry);
		if (!p)
			return pe;
	}

	p = ScanTuple(p, 0, entry, ps);
	if (!p)
		return pe;

	/* [UNREPLIED] taken care of in ScanTuple */

	p = ScanTuple(p, 1, entry, ps);
	if (!p)
		return pe;

	while (*p) {
		SKIP_BLANK(p)
		if (!strncmp(p, "[ASSURED]", 9)) {
			p += 9;
			entry->flags |= F_ASSURED;
		}
		else
			SKIP_NON_BLANK(p)
	}

	*err = 0;
	return pe;
}

char *
ScanTcp(char *p, struct ipConntrackEntry *entry)
{
	char *s;
	int n, l;

	SKIP_BLANK(p)

	for (n = 0; tcp_states[n].lName; ++n) {
		s = tcp_states[n].lName;
		l = strlen(s);
		if (!strncmp(p, s, l) && p[l] == ' ') {
			entry->u.tcp_udp.state = n;
			return p + l + 1;
		}
	}

	return p;
}

char *
ScanTuple(char *p, int n, struct ipConntrackEntry *entry, char *ps)
{
	SKIP_BLANK(p)
	while (*p && strncmp("src=", p, 4)) {
		if (!strncmp(p, "[UNREPLIED] ", 12)) {
			p += 12;
			entry->flags |= F_UNREPLIED;
		}
		else
			SKIP_NON_BLANK(p)
		SKIP_BLANK(p)
	}
	if (!*p) {
		Log(LOG_ERROR, 0, "scan error 10: %s", ps);
		return NULL;
	}
	p += 4;
	if (!inet_aton(p, &entry->src[n])) {
		Log(LOG_ERROR, 0, "scan error 11: %s", ps);
		return NULL;
	}
	SKIP_NON_BLANK(p)
	SKIP_BLANK(p)
	if (strncmp("dst=", p, 4)) {
		Log(LOG_ERROR, 0, "scan error 12: %s", ps);
		return NULL;
	}
	p += 4;
	if (!inet_aton(p, &entry->dst[n])) {
		Log(LOG_ERROR, 0, "scan error 13: %s", ps);
		return NULL;
	}
	SKIP_NON_BLANK(p)
	SKIP_BLANK(p)

	if (entry->protocol == IPPROTO_TCP || entry->protocol == IPPROTO_UDP)
		p = ScanTupleTcpUdp(p, n, entry, ps);
	else if (entry->protocol == IPPROTO_ICMP)
		p = ScanTupleIcmp(p, n, entry, ps);

	return p;
}

char *
ScanTupleTcpUdp(char *p, int n, struct ipConntrackEntry *entry, char *ps)
{
	int l;

	SKIP_BLANK(p)
	while (*p && strncmp("sport=", p, 6)) {
		SKIP_NON_BLANK(p)
		SKIP_BLANK(p)
	}

	if (sscanf(p, "sport=%hu dport=%hu %n", &entry->u.tcp_udp.sport[n],
					&entry->u.tcp_udp.dport[n], &l) != 2) {
		Log(LOG_ERROR, 0, "scan error 20: %s", ps);
		return NULL;
	}

	return p + l;
}

char *
ScanTupleIcmp(char *p, int n, struct ipConntrackEntry *entry, char *ps)
{
	int l;
	unsigned type, code, id;

	SKIP_BLANK(p)
	while (*p && strncmp("type=", p, 5)) {
		SKIP_NON_BLANK(p)
		SKIP_BLANK(p)
	}

	if (sscanf(p, "type=%u code=%u id=%u %n", &type, &code, &id, &l) != 3) {
		Log(LOG_ERROR, 0, "scan error 30: %s", ps);
		return NULL;
	}
	entry->u.icmp.type[n] = type;
	entry->u.icmp.code[n] = code;
	entry->u.icmp.id[n] = id;

	return p + l;
}

struct ipConntrackEntry *
AllocEntry()
{
	struct ipConntrackEntry *p;

	if (entryCache) {
		p = entryCache;
		entryCache = entryCache->next;
		return p;
	}

	return malloc(sizeof(struct ipConntrackEntry));
}

void
FreeEntry(struct ipConntrackEntry *e)
{
	e->next = entryCache;
	entryCache = e;
}

unsigned 
Hash(struct ipConntrackEntry *e)
{
	unsigned h;

	h = e->protocol ^
	    e->src[0].s_addr ^ e->src[1].s_addr;
	if (e->protocol == IPPROTO_TCP || e->protocol == IPPROTO_UDP)
		h ^= e->u.tcp_udp.sport[0] ^ e->u.tcp_udp.sport[1];
	else if (e->protocol == IPPROTO_ICMP)
		h ^= e->u.icmp.type[0] ^ e->u.icmp.code[0] ^ e->u.icmp.id[0];
	
	return h % MAX_CONN_HASH;
}

int 
Equal(struct ipConntrackEntry *e1, struct ipConntrackEntry *e2)
{
	if (e1->protocol != e2->protocol)
		return 0;

	if (e1->src[0].s_addr != e2->src[0].s_addr ||
	    e1->src[1].s_addr != e2->src[1].s_addr)
	    	return 0;

	if (e1->protocol == IPPROTO_TCP || e1->protocol == IPPROTO_UDP) {
		if (e1->u.tcp_udp.sport[0] != e2->u.tcp_udp.sport[0] ||
		    e1->u.tcp_udp.sport[1] != e2->u.tcp_udp.sport[1])
		    	return 0;
	}
	else if (e1->protocol == IPPROTO_ICMP) {
		if (e1->u.icmp.type[0] != e2->u.icmp.type[0] ||
		    e1->u.icmp.code[0] != e2->u.icmp.code[0] ||
		    e1->u.icmp.id[0] != e2->u.icmp.id[0])
		    	return 0;
	}

	return 1;
}

struct ipConntrackEntry *
FindInsert(struct ipConntrackEntry *e)
{
	unsigned h;
	struct ipConntrackEntry *p;

	h = Hash(e);
	for (p = connHash[h]; p && !Equal(p, e); p = p->next)
		;
	if (p)
		return p;

	e->next = connHash[h];
	connHash[h] = e;
	return e;
}

void
SetTermFlag(int sig)
{
	termFlag = 1;
}

void 
Log(char *type, time_t t, char *msg, ...)
{
	va_list ap;
#       define LOG_TIME_STR_LEN 40
	char s[LOG_TIME_STR_LEN];
	struct tm *tm;

	if (!strcmp(type, LOG_ERROR))
		++errCnt;

	if (!t)
		t = time(NULL);

	tm = localtime(&t);
	strftime(s, LOG_TIME_STR_LEN, "%F-%T", tm);
	va_start(ap, msg);
	fprintf(outFp, "%s %s ", type, s);
	vfprintf(outFp, msg, ap);
	fprintf(outFp, "\n");
	va_end(ap);
}

void
PrintEntry(struct ipConntrackEntry *entry)
{
	printf("protocol %u, timeout %u, src ", 
		entry->protocol, entry->timeout);
	printf("%s, ", inet_ntoa(entry->src[0]));
	printf("%s, dst ", inet_ntoa(entry->src[1]));
	printf("%s, ", inet_ntoa(entry->dst[0]));
	printf("%s, ", inet_ntoa(entry->dst[1]));
	printf("flags 0x%x, ", entry->flags);
	if (entry->protocol == IPPROTO_TCP || 
	    entry->protocol == IPPROTO_UDP) {
		printf("state %u, sport %u, %u, dport %u, %u",
			entry->u.tcp_udp.state, 
			entry->u.tcp_udp.sport[0], 
			entry->u.tcp_udp.sport[1],
			entry->u.tcp_udp.dport[0], 
			entry->u.tcp_udp.dport[1]);
	} else if (entry->protocol == IPPROTO_ICMP) {
		printf("type %u, %u, code %u, %u, id %u, %u",
			entry->u.icmp.type[0],
			entry->u.icmp.type[1],
			entry->u.icmp.code[0],
			entry->u.icmp.code[1],
			entry->u.icmp.id[0],
			entry->u.icmp.id[1]);
	}
	printf("\n");
}

void 
LogEntry(char *type, time_t t, struct ipConntrackEntry *entry)
{
#	define ADDR_LEN 20
	char src0[ADDR_LEN], src1[ADDR_LEN];
	char snat_addr[ADDR_LEN], dnat_addr[ADDR_LEN];
	char prot[8], f[8], sport_type[8], dport_code[8];
	char snat_port[8], dnat_port[8];
	char tim[20] = "-";
	int is_snat = 0, is_dnat = 0;

	if (entry->intFlags & FI_IGNORE)
		return;
		
	if (entry->src[0].s_addr != entry->dst[1].s_addr)
		is_snat = 1;
	if (entry->src[1].s_addr != entry->dst[0].s_addr)
		is_dnat = 1;

	*f = 0;
	if (entry->flags & F_UNREPLIED)
		strcpy(f, "U");

	if (entry->protocol == IPPROTO_TCP || entry->protocol == IPPROTO_UDP){
		if (entry->protocol == IPPROTO_TCP) {
			strcpy(prot, "tcp");
			strcat(f, tcp_states[entry->u.tcp_udp.state].sName);
		}
		else
			strcpy(prot, "udp");

		sprintf(sport_type, "%u", entry->u.tcp_udp.sport[0]);
		sprintf(dport_code, "%u", entry->u.tcp_udp.sport[1]);

		if (is_snat)
			sprintf(snat_port, "%u", entry->u.tcp_udp.dport[1]);
		else
			strcpy(snat_port, "-");
		if (is_dnat)
			sprintf(dnat_port, "%u", entry->u.tcp_udp.dport[0]);
		else
			strcpy(dnat_port, "-");
	}
	else if (entry->protocol == IPPROTO_ICMP) {
		strcpy(prot, "icmp");
		sprintf(sport_type, "%u", entry->u.icmp.type[0]);
		sprintf(dport_code, "%u", entry->u.icmp.code[0]);
		strcpy(snat_port, "-");
		strcpy(dnat_port, "-");
	}
	else {
		sprintf(prot, "%u", entry->protocol);
		strcpy(sport_type, "-");
		strcpy(dport_code, "-");
		strcpy(snat_port, "-");
		strcpy(dnat_port, "-");
	}

	if (!*f)
		strcpy(f, "-");

	AddrToAscii(src0, sizeof(src0), entry->src[0], entry->srcIfIdx[0]);
	AddrToAscii(src1, sizeof(src1), entry->src[1], entry->srcIfIdx[1]);

	if (is_snat)
		AddrToAscii(snat_addr, sizeof(snat_addr), 
			entry->dst[1], entry->dstIfIdx[1]);
	else
		strcpy(snat_addr, "-");
	if (is_dnat)
		AddrToAscii(dnat_addr, sizeof(dnat_addr),
			entry->dst[0], entry->dstIfIdx[0]);
	else
		strcpy(dnat_addr, "-");

	if (type == LOG_CLOSE_CONN || type == LOG_EXIT_CONN || 
						type == LOG_TCP_STATE)
		sprintf(tim, "%5u", t - entry->openTime);

	if (nFlag)
	    Log(type, t, "%-4s %-15s %-15s %-5s %-5s %-2s %5s "
	    	"%-15s %-15s %5s %5s",
		prot, src0, src1, sport_type, dport_code, f, tim,
		snat_addr, dnat_addr, snat_port, dnat_port);
	else
	    Log(type, t, "%-4s %-15s %-15s %-5s %-5s %-2s %5s",
		prot, src0, src1, sport_type, dport_code, f, tim);
}

void 
ParseCfgFile(char *cfgfile)
{
	FILE *fp;
	char buf[512];
	char *p, *iF, *newIf, *addr, *m;
	int n, val;

	fp = fopen(cfgfile, "r");
	if (!fp) {
		*cfgfile = 0;
		return;
	}

	n = 0;
	while (fgets(buf, sizeof(buf), fp)) {
		++n;
		p = strtok(buf, " \t\n");
		if (!p || *p == '#')
			continue;
		if (!strcmp(p, "debug"))
			dFlag = 1;
		else if (!strcmp(p, "ifname")) {
			if (ifMapLen >= IF_MAP_MAX)
			    	Error("%s, line %d: too many ifname directives",
			    		cfgfile, n);

			iF = strtok(NULL, " \t\n");
			if (!iF)
				Error("%s, line %d: bad ifname directive", 
					cfgfile, n);
			if (strlen(iF) > IF_NAMESIZE)
				Error("%s, line %d: ifname too long", 
					cfgfile, n);
			strncpy(ifMap[ifMapLen].name, iF, IF_NAMESIZE);
			ifMap[ifMapLen].name[IF_NAMESIZE] = 0;

			addr = strtok(NULL, " \t\n");
			if (!addr)
				Error("%s, line %d: bad ifname directive", 
					cfgfile, n);
			m = index(addr, '/');
			if (!m)
				ifMap[ifMapLen].mask = ~0;
			else {
				*m++ = 0;
				val = atoi(m);
				if (val < 1 || val > 32)
					Error("%s, line %d: bad address %s/%s",
						cfgfile, n, addr, m);
				ifMap[ifMapLen].mask = 
					ntohl(~( (1 << (32 - val)) - 1));
			}
			if (!inet_aton(addr, &ifMap[ifMapLen].addr))
				Error("%s, line %d: bad address %s/%s", 
					cfgfile, n, addr, m);

			newIf = strtok(NULL, " \t\n");
			if (!newIf)
				Error("%s, line %d: bad ifname directive", 
					cfgfile, n);
			if (strlen(newIf) > IF_NAMESIZE)
				Error("%s, line %d: ifname too long", 
					cfgfile, n);
			strncpy(ifMap[ifMapLen].newName, newIf, IF_NAMESIZE);
			ifMap[ifMapLen].newName[IF_NAMESIZE] = 0;

			++ifMapLen;
		}
		else
			Error("%s, line %d: unknown directive %s",
				cfgfile, n, p);
	}

	fclose(fp);
}

/*
 * If interface name "name" matches a description in ifMap replace it
 * with ifMap[].newName.
 */

void 
RemapIf(char *name, struct in_addr addr)
{
	int n;
	unsigned m;

	for (n = 0; n < ifMapLen; ++n) {
		if (!strcmp(ifMap[n].name, name)) {
			m = ifMap[n].mask;
			if ((ifMap[n].addr.s_addr & m) == (addr.s_addr & m)) {
				strcpy(name, ifMap[n].newName);
				return;
			}
		}
	}
}

void
Usage(char *prog)
{
	errno = 0;
	Error(
		"usage: %s [options]\n"
		"    options:\n"
		"       -d               : run in foreground\n"
		"       -h               : help\n"
		"       -o output-file   : send log output to output-file\n"
		"       -t time-interval : for ip_conntrack lookups (secs)\n"
		"       -I               : show icmp entries\n"
		"       -i               : don't show if name instead of addr\n"
		"       -N               : show NAT info\n"
		"       -1               : show open connections and exit\n"
				, prog);
}

void
Error(char *str, ...)
{
        va_list ap;
	int e;
	char buf[512], *p;

	e = errno;
        va_start(ap, str);

	if (syslog_open) {
		vsnprintf(buf, sizeof(buf), str, ap);
		p = buf + strlen(buf);
		if (e) {
			snprintf(p, buf + sizeof(buf) - p,
				", errno %d: %s", e, strerror(e));
			p += strlen(p);
		}
		snprintf(p, buf + sizeof(buf) - p, "\n" );
		p += strlen(p);
		syslog(LOG_ERR, buf);
	}
	else {
        	vfprintf(stderr, str, ap);
        	fprintf(stderr, "\n");
		if (e)
			fprintf(stderr, "   errno %d: %s\n", e, strerror(e));
	}

	va_end(ap);

        exit(1);
}
